/* Copyright (C) 2012 Kristian Lauszus, TKJ Electronics. All rights reserved.

   This software may be distributed and modified under the terms of the GNU
   General Public License version 2 (GPL2) as published by the Free Software
   Foundation and appearing in the file GPL2.TXT included in the packaging of
   this file. Please note that GPL2 Section 2[b] requires that all works based
   on this software must also be made publicly available under the terms of
   the GPL2 ("Copyleft").

   Contact information
   -------------------

   Kristian Lauszus, TKJ Electronics
Web      :  http://www.tkjelectronics.com
e-mail   :  kristianl@tkjelectronics.com
*/

#include "SPP.h"
#include "BTD.h"

#define DEBUG_USB_HOST

#ifdef DEBUG_USB_HOST
#undef TRACE_USBHOST(x)
#define TRACE_USBHOST(x)
#endif

SPP::SPP(BTD *p) : pBtd(p), service_class_id(0) // Pointer to BTD class instance - mandatory
{/*{{{*/
	timer = GetTickCount();
}/*}}}*/

void SPP::Initialize(BTD *p) {/*{{{*/
	pBtd = p;
	if (pBtd) {

    // Register it as a Bluetooth service
		service_class_id = pBtd->registerServiceClass((BluetoothService *)this);

		TRACE_USBHOST(printf("-> Registered SPP as service #%u\r\n", service_class_id);)

		/* Set device cid for the SDP and RFCOMM channels */
    *(uint16_t *)(&sdp_dcid)    = 0x0050;
    *(uint16_t *)(&rfcomm_dcid) = 0x0051;

		Reset();
    UHD_Init();
	}
}/*}}}*/

SPP::operator bool() {/*{{{*/
	return connected;
}/*}}}*/

void SPP::Reset() {/*{{{*/
	connected          = false;
	RFCOMMConnected    = false;
	SDPConnected       = false;
	waitForLastCommand = false;
	l2cap_sdp_state    = L2CAP_SDP_WAIT;
	l2cap_rfcomm_state = L2CAP_RFCOMM_WAIT;
	l2cap_event_flag   = 0;
	spp_available_out  = 0;
}/*}}}*/

void SPP::disconnect() {/*{{{*/
	connected = false;
	// First the two L2CAP channels has to be disconnected and then the HCI connection
	if (RFCOMMConnected)
		pBtd->enqueue_l2cap_disconnection_request(getServiceId(), hci_handle, 0x0A, rfcomm_scid, rfcomm_dcid);
	if (SDPConnected)
		pBtd->enqueue_l2cap_disconnection_request(getServiceId(), hci_handle, 0x0B, sdp_scid, sdp_dcid);
	l2cap_sdp_state = L2CAP_DISCONNECT_RESPONSE;
}/*}}}*/

uint8_t SPP::getServiceId() {/*{{{*/
	return service_class_id;
}/*}}}*/

void dump_l2cap_message_t(l2cap_message_t * p) {
  printf("	handle:\t0x%04.4X\r\n"          , p->handle);           // 0-1
  printf("	unused_1[2]:\t0x%04.4X\r\n"     , *((uint16_t *)&p->unused_1));      // 2-5
  printf("	l2cap_length:\t0x%04.4X\r\n"    , p->l2cap_length);       // 6-7
  printf("	channel_id:\t0x%04.4X\r\n"      , p->channel_id);       // 6-7
  printf("	msgtype:\t0x%02.2X\r\n"         , p->msgtype);          // 8
  printf("	identifier:\t0x%02.2X\r\n"      , p->identifier);       // 9
  printf("	unused_2[2]:\t0x%04.4X\r\n"     , *((uint16_t *)&p->unused_2));      // 10-11
  printf("	psm:\t0x%04.4X\r\n"             , p->psm);              // 12-13
  printf("	service_class_id:\t0x%04.4X\r\n", p->service_class_id); // 14-15
  printf("	response:\t0x%04.4X (flip 0x%04.4X)\r\n"        , p->response, byteflip(p->response)); // 16-17
  printf("	uuid_high:\t0x%04.4X\r\r\n"     , p->uuid_high); // 18-19
}

bool SPP::handle_signaling_channel(uint8_t *l2capinbuf)
{/*{{{*/

  l2cap_message_t * p = (l2cap_message_t *)l2capinbuf;
  bool handle_is_active = p->handle == (hci_handle | 0x2000);
  bool processing_done  = true;

  switch ( p->msgtype ) {/*{{{*/

    case L2CAP_CMD_COMMAND_REJECT: 

      TRACE_USBHOST(printf("%lu L2CAP Command Rejected - Reason: 0x%04.4X", GetTickCount(),p->psm);)
      TRACE_USBHOST(printf(" Response: 0x%04.4X;  Service Class ID: %04.4X\r\n", p->response, p->service_class_id);)
      break;

    case L2CAP_CMD_CONNECTION_REQUEST: 
      // We have received an L2CAP connection request packet
      if (!connected)
      {/*{{{*/
        switch ( p->psm ) {/*{{{*/
          case SDP_PSM:
            if ( !pBtd->sdpConnectionClaimed ) {
              pBtd->sdpConnectionClaimed = true;
              hci_handle         = pBtd->hci_handle; // Store the HCI Handle for the connection
              l2cap_sdp_state    = L2CAP_SDP_WAIT;   // Reset state
            }
            break;
          case RFCOMM_PSM:
            if (!pBtd->rfcommConnectionClaimed) {
              pBtd->rfcommConnectionClaimed = true;
              hci_handle         = pBtd->hci_handle;  // Store the HCI Handle for the connection
              l2cap_rfcomm_state = L2CAP_RFCOMM_WAIT; // Reset state
            }
            break;
          default:
            printf("%lu !! Unknown proto/serv multiplexer (PSM) value 0x%04.4X\r\n", GetTickCount(), p->psm);
            dump_l2cap_message_t(p);
            memset(l2capinbuf,0,L2CAP_IN_BUFFERSIZE);
            return false;
            break;
        }/*}}}*/
      }/*}}}*/

      TRACE_USBHOST(printf("%lu ~~ L2CAP Connection Request - PSM: 0x%04.4X", GetTickCount(), p->psm);)
      TRACE_USBHOST(printf(" SCID: 0x%04.4X", p->service_class_id);)
      TRACE_USBHOST(printf(" Identifier: 0x%02.2x\r\n", p->identifier);)

      switch ( p->psm ) {/*{{{*/
        case SDP_PSM:
          // Retain transaction identifier (?)
          identifier = p->identifier;
          // Keep the source CID as this host's destination CID
          memcpy((uint8_t *)&sdp_scid, &p->service_class_id, 2);
          l2cap_sdp_state =  L2CAP_SDP_WAIT;
          l2cap_event_flag &= ~L2CAP_FLAG_CONNECTION_SDP_REQUEST; // Clear flag
          if ( handle_is_active ) {
            TRACE_USBHOST(printf("%lu ~~ Enqueueing SDP Response - PSM: 0x%04.4X\r\n", GetTickCount(), p->psm);)
            pBtd->enqueue_l2cap_message(getServiceId(), L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR, sdp_scid, p->identifier, l2capinbuf[10] );
          }
          else {
            TRACE_USBHOST(printf("%lu ~~ Enqueueing Connection Response - PSM: 0x%04.4X\r\n", GetTickCount(), p->psm);)
            pBtd->enqueue_l2cap_connection_response(getServiceId(), hci_handle, identifier, sdp_dcid, sdp_scid, PENDING);
            pBtd->enqueue_l2cap_connection_response(getServiceId(), hci_handle, identifier, sdp_dcid, sdp_scid, SUCCESSFUL, SDP_CONN_INCREMENT_CONNID);
          }
          break;
        case RFCOMM_PSM:
          // ----- || -----
          TRACE_USBHOST(printf("%lu ~~ Enqueueing RFCOMM Response - PSM: 0x%04.4X\r\n", GetTickCount(), p->psm);)
          identifier = p->identifier;
          memcpy((uint8_t *)&rfcomm_scid, &p->service_class_id, 2);
          if ( l2cap_rfcomm_state == L2CAP_RFCOMM_WAIT ) {
            l2cap_rfcomm_state = L2CAP_RFCOMM_HOLD; 
            l2cap_event_flag &= ~L2CAP_FLAG_CONFIG_RFCOMM_REQUEST;
            pBtd->enqueue_l2cap_connection_response(getServiceId(), hci_handle, identifier, rfcomm_dcid, rfcomm_scid, PENDING);
            pBtd->enqueue_l2cap_connection_response(getServiceId(), hci_handle, identifier, rfcomm_dcid, rfcomm_scid, SUCCESSFUL, RFCOMM_CONN_INCREMENT_CONNID);
          }
          break;
        default:
          printf("%lu ? Unknown L2CAP_CMD_CONNECTION_REQUEST PSM 0x%04.4X\r\n", GetTickCount(), p->psm);
          printf("%lu !! Information request to unknown PSM 0x%04.4X\r\n", GetTickCount(), p->psm);
          dump_l2cap_message_t(p);
          return false;
          break;
      }/*}}}*/

      break;

    case L2CAP_CMD_CONFIG_RESPONSE:
      if (p->response != 0x0000) {
        TRACE_USBHOST(printf("%lu ? L2CAP_CMD_CONFIG_RESPONSE failure code 0x%04.4X\r\n", GetTickCount(), p->response);)
      } 
      else if (p->psm == *((uint16_t *)&sdp_dcid)) {
        TRACE_USBHOST(printf("%lu Recv L2CAP Configuration Response, SDP\r\n", GetTickCount());)
				firstMessage = true; // Reset bool
				SDPConnected = true;
				l2cap_sdp_state = L2CAP_SDP_DONE;
      }
      else if (p->psm == *((uint16_t *)&rfcomm_dcid)) {
        TRACE_USBHOST(printf("%lu Recv L2CAP Configuration Response, RFCOMM\r\n", GetTickCount());)
        if ( l2cap_rfcomm_state == L2CAP_RFCOMM_REQUEST ) {
          l2cap_rfcomm_state = L2CAP_RFCOMM_HOLD;
          // Echo config response; update state machine only after command is 
          // dispatched (IHC sets state to L2CAP_RFCOMM_DONE)
        }
      }
      else {
        printf("%lu ? Unknown L2CAP_CMD_CONFIG_RESPONSE PSM 0x%04.4X\r\n", GetTickCount(), p->psm);
      }

      break;

    case L2CAP_CMD_CONFIG_REQUEST: 

      if (p->psm == *((uint16_t *)&sdp_dcid)) {
        TRACE_USBHOST(printf("%lu ~~ L2CAP Configuration Request for SDP, PSM 0x%04.4X\r\n", GetTickCount(), p->psm);)
        identifier = p->identifier;
        // L2CAP_CONFIG_RESPONSE
        pBtd->enqueue_l2cap_config_response(getServiceId(), hci_handle, identifier, sdp_scid, L2CAP_CONFIG_RESPONSE_SENT);
        // Enqueue an SDP configuration request, following which this host should 
        // receive SDP Service Search Attribute Requests.  
        pBtd->enqueue_l2cap_config_request(getServiceId(), hci_handle, identifier, sdp_scid, L2CAP_CONFIG_REQUEST_SENT);
        // pBtd->enqueue_l2cap_message(getServiceId(), L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR, sdp_scid, p->identifier, l2capinbuf[10] );
      }
      else if (p->psm == *((uint16_t *)&rfcomm_dcid)) {
        TRACE_USBHOST(printf("%lu ~~ L2CAP Configuration Request for RFCOMM, PSM 0x%04.4X\r\n", GetTickCount(), p->psm);)
        identifier = p->identifier;

        pBtd->enqueue_l2cap_config_response(getServiceId(), hci_handle,
          identifier, rfcomm_scid, RFCOMM_CONN_RECEIVED_CONFIG_REPLY);
				pBtd->enqueue_l2cap_config_request(getServiceId(), hci_handle,
          identifier, rfcomm_scid, RFCOMM_CONN_L2CAP_RFCOMM_REQUEST);
      }
      else {
        TRACE_USBHOST(printf("%lu ? Unknown L2CAP_CMD_CONFIG_REQUEST PSM 0x%04.4X\r\n", GetTickCount(), p->psm);)
      }

      break;

    case L2CAP_CMD_DISCONNECT_REQUEST: 
      if (p->psm == *((uint16_t *)&sdp_dcid)) {/*{{{*/

        TRACE_USBHOST(printf("%lu Disconnect Request: SDP Channel\r\n", GetTickCount());)

				l2cap_event_flag &= ~L2CAP_FLAG_DISCONNECT_SDP_REQUEST; // Clear flag

        identifier = p->identifier;
				pBtd->enqueue_l2cap_disconnection_response(
          getServiceId(),
          hci_handle,
          identifier,
          sdp_dcid,
          sdp_scid, SDP_CONN_DISCONNECT_SDP_REQUEST
        );
      }/*}}}*/
      else if (p->psm == *((uint16_t *)&rfcomm_dcid)) {/*{{{*/

        TRACE_USBHOST(printf("%lu Disconnect Request: RFCOMM Channel\r\n", GetTickCount());)

				l2cap_event_flag &= ~L2CAP_FLAG_DISCONNECT_RFCOMM_REQUEST; // Clear flag

        identifier = p->identifier;
				pBtd->enqueue_l2cap_disconnection_response(
          getServiceId(),
          hci_handle,
          identifier,
          rfcomm_dcid,
          rfcomm_scid, RFCOMM_CONN_DISCONNECT_REQUEST
        );
      }/*}}}*/
      else {
        printf("%lu ? Unknown L2CAP_CMD_DISCONNECT_REQUEST PSM 0x%04.4X\r\n", GetTickCount(), p->psm);
      }

      break;

    case L2CAP_CMD_DISCONNECT_RESPONSE: 

      if (p->psm == *((uint16_t *)&sdp_scid)) {
        TRACE_USBHOST(printf("%lu Disconnect Response: SDP Channel\r\n", GetTickCount());)
        identifier = p->identifier;
      }
      else if (p->psm == *((uint16_t *)&rfcomm_scid)) {
        TRACE_USBHOST(printf("%lu Disconnect Response: RFCOMM Channel\r\n", GetTickCount());)
        identifier = p->identifier;
      }
      else {
        printf("%lu ? Disconnect response with unknown PSM 0x%04.4X\r\n", GetTickCount(), p->psm);
        break;
      }

      pBtd->hci_disconnect(hci_handle);

      RFCOMMConnected               = false;
      SDPConnected                  = false;
      hci_handle                    = -1; // Reset handle
      l2cap_event_flag              = 0; // Reset flags
      l2cap_sdp_state               = L2CAP_SDP_WAIT;
      l2cap_rfcomm_state            = L2CAP_RFCOMM_WAIT;
      pBtd->sdpConnectionClaimed    = false;
      pBtd->rfcommConnectionClaimed = false;

      break;

    case L2CAP_CMD_INFORMATION_REQUEST: 
      TRACE_USBHOST(printf("%lu Information request\r\n", GetTickCount());)
      identifier = p->identifier;
      pBtd->enqueue_l2cap_message_initiator(
        getServiceId(),
        L2CAP_QUEUE_INFORMATION_RESPONSE,
        hci_handle,
        identifier,
        p->psm & 0xFF,
        (p->psm >> 8) & 0xFF
      );

      break;

    default:
      if ( p->msgtype != 0 ) {
        TRACE_USBHOST(printf("%lu !! %c%c%c%c L2CAP Unknown Signaling Command: 0x%02.2x\r\n", GetTickCount(),p->msgtype);)
        dump_l2cap_message_t(p);
      }
      else processing_done = false;

  }/*}}}*/

  if ( processing_done ) pBtd->hci_host_number_of_packets_completed(1);

  return true;

}/*}}}*/

const char * sdp_pdu_id_names[] = {
  "{Not SDP PDU ID}",
  "ERROR_RESPONSE",
  "SERVICE_SEARCH_REQUEST",
  "SERVICE_SEARCH_RESPONSE",
  "SERVICE_ATTRIBUTE_REQUEST",
  "SERVICE_ATTRIBUTE_RESPONSE",
  "SERVICE_SEARCH_ATTRIBUTE_REQUEST",
  "SERVICE_SEARCH_ATTRIBUTE_RESPONSE",
};

bool SPP::handle_sdp_channel(uint8_t *l2capinbuf)
{/*{{{*/

  l2cap_message_t * p   = (l2cap_message_t *)l2capinbuf;
  bool handle_is_active = p->handle == (hci_handle | 0x2000);
  sdp_pdu_header_t * sh = (sdp_pdu_header_t *)((uint8_t *)(&p->msgtype));
  uint8_t sp_counter    = 0;
  uint8_t * des_token_pos = &(sh->start_of_data);
  bool processing_done  = true;

  sdp_data_element_t sdp, de;

  printf("%lu {0x%02.2X} SDP_%s, TID 0x%04.4X length 0x%04.4X\r\n",
    GetTickCount(),
    sh->pdu_id,
    sdp_pdu_id_names[sh->pdu_id <= SDP_PDU_ID_MAX ? sh->pdu_id : 0],
    byteflip(sh->transaction_id),
    byteflip(sh->parameter_length)
  );
  /*
   * 15063 {0x06} SDP_SERVICE_SEARCH_ATTRIBUTE_REQUEST, TID 0x0001 length 0x000F
   * ~~ Public Browse Root Request
   * ** PDU element DATASEQ (0x06) type size 0x05 length 3 (0x35) next 6
   *      0x10 0x02 0xff!! Bad SDP PDU type desc 31
   *
   *      15063   27 20 18 00 14 00 50 00  06 00 01 00 0f 35 03 19
   *      15063   10 02 ff ff 35 05 0a 00  00 ff ff 00 00 00 00 00
   *      15063   00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00
   *      15063   00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00
   */
  switch ( sh->pdu_id ) {/*{{{*/

    case SDP_ERROR_RESPONSE_PDU : // 0x01
      break;

    case SDP_SERVICE_SEARCH_REQUEST_PDU : // 0x02
      break;

    case SDP_SERVICE_SEARCH_RESPONSE_PDU : // 0x03
      break;

    case SDP_SERVICE_ATTRIBUTE_REQUEST_PDU : // 0x04
      break;

    case SDP_SERVICE_ATTRIBUTE_RESPONSE_PDU : // 0x05
      break;

    case SDP_SERVICE_SEARCH_ATTRIBUTE_REQUEST_PDU : // 0x06 // See the RFCOMM specs
      {/*{{{*/

        uint16_t request_uuid = 0xFFFF;
        bool testing_128bit_uuid = false;

        if ( pBtd->data_element_sequence_tokenizer( &sdp, &des_token_pos ) ) {/*{{{*/

          // The request will contain, in order:
          // - A Data Sequence element of up to 12 UUIDs;
          // - A MaximumAttributeByteCount parameter (2 bytes);
          // - An AttributeIDList (variable length), either a
          //   - Sequence of 16-bit attribute UUIDs, or
          //   - A 32-bit attribute UUID range
          
          // Take the data element UUID as our request_uuid
          if ( pBtd->data_element_sequence_tokenizer( &de, &(sdp.data) ) ) {
            request_uuid = byteflip(*((uint16_t *)de.data));
#ifdef EXTRADEBUG
            printf("> Parsed request UUID 0x%04.4X\r\n", request_uuid);
#endif
          }
#ifdef EXTRADEBUG
          // Skip past the MaximumAttributeByteCount word
          des_token_pos = &(des_token_pos[2]);
          // Parse the AttributeIDList
          if ( pBtd->data_element_sequence_tokenizer( &sdp, &des_token_pos ) ) {
            printf("> Parsed AttributeIDList 0x%08.8X\r\n", ((uint32_t *)&(sdp.data[0])));
          }
#endif

        }/*}}}*/

test_128bit_uuid:

        switch ( request_uuid ) {/*{{{*/
          case SDP_PublicBrowseRoot:
            printf("~~ Public Browse Root Request\r\n");
            pBtd->enqueue_l2cap_message(getServiceId(), L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR, sdp_scid, l2capinbuf[10], p->identifier );
            break;                                                                                                                       
          case SERIALPORT_UUID:                                                                                                          
            printf("~~ Serial Port SDP SAR\r\n");
            pBtd->enqueue_l2cap_message(getServiceId(), L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR, sdp_scid, l2capinbuf[10], p->identifier );
            break;
          case L2CAP_UUID:
            printf("~~ L2CAP SDP SAR\r\n");
            pBtd->enqueue_l2cap_message(getServiceId(), L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR, sdp_scid, l2capinbuf[10], p->identifier );
            break;
          case 0x0000:
            if ( byteflip(p->uuid_high) == 0x0000 ) {/*{{{*/
              // Connectionless SDP record browse?
              TRACE_USBHOST(printf("%lu ~~ Connectionless SDP service search attribute request\r\n", GetTickCount());)
              dump_l2cap_message_t(p);
              l2cap_sdp_state = L2CAP_SDP_WAIT;
              pBtd->enqueue_l2cap_message(getServiceId(), L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR, sdp_scid, l2capinbuf[10], p->identifier, SDP_DISCOVERY_CONNECTIONLESS );
              break;
            }/*}}}*/
            else if (!testing_128bit_uuid) {/*{{{*/
              testing_128bit_uuid = true;
              request_uuid = byteflip(p->uuid_high);
              goto test_128bit_uuid;
            }/*}}}*/
            // Fallthrough is deprecated in MISRA C.
            // Hobbyists will want to modify this. 
          default:
            pBtd->enqueue_l2cap_message(getServiceId(), L2CAP_SERVICE_NOT_SUPPORTED, sdp_scid, l2capinbuf[10], p->identifier); // The service is not supported
            printf("%lu Unsupported UUID 0x%04.4X (0x%04.4X) against either SERIALPORT_UUID (0x%04.4X) or L2CAP_UUID (0x%04.4X)\r\n", GetTickCount(),
              request_uuid,
              byteflip(p->response),
              SERIALPORT_UUID,
              L2CAP_UUID
            );
#ifdef EXTRADEBUG
            for (uint16_t i = 0; i < L2CAP_IN_BUFFERSIZE; i++) {
              if (!(i % 16)) printf("\r\n%lu ", GetTickCount());
              if (!(i % 8)) printf(" ");
              printf(" %02.2x", l2capinbuf[i]);
            }
            printf("\r\n");

            dump_l2cap_message_t(p);
#endif
            break;
        }/*}}}*/

      }/*}}}*/
      break;

    case SDP_SERVICE_SEARCH_ATTRIBUTE_RESPONSE_PDU : // 0x07 // See the RFCOMM specs
      break;

    default:
      processing_done = false;

  }/*}}}*/

  if ( processing_done ) pBtd->hci_host_number_of_packets_completed(1);

  return true;

}/*}}}*/

#undef TRACE_USBHOST(x)
#define TRACE_USBHOST(x) x

bool SPP::handle_rfcomm_channel(uint8_t *l2capinbuf)
{/*{{{*/

  l2cap_message_t * p = (l2cap_message_t *)l2capinbuf;
  l2cap_rfcomm_packet_t * r = (l2cap_rfcomm_packet_t *)l2capinbuf;

  bool dump_frame = false;

  // RFCOMM Spec, see sec. 5.4, DLCI Allocation with RFCOMM Server Channels
  // Note: You need to support multiple DLCIs to be able to provide access to
  // all Due TX/RX UART serial ports.
  // (a DLCI is identified by combining rfcommChannel | rfcommDirection)
  //
  // There is exactly one RFCOMM frame per L2CAP frame
  //
  // Control channel: DLCI 0
  rfcommChannel         = r->address & 0xF8; // b 1111 1000 - Channel 1-30 (DLCIs 0, 31 reserved by GSM 7.10)
  rfcommDirection       = r->address & 0x04; // b 0000 0100 - Direction rel to initiator 
  rfcommCommandResponse = r->address & 0x02; // b 0000 0010 - Command (else response)
  rfcommFrameType       = r->control & 0xEF; // b 1110 1111 - Control Field, See 5.2.1.3, TS 07.10
  rfcommPfBit           = r->control & 0x10; // b 0001 0000 - Poll/Final Bit

  // Why?
  if (rfcommChannel >> 3 != 0x00)
    rfcommChannelConnection = rfcommChannel;

  if ( rfcommFrameType != 0xef )
  TRACE_USBHOST(printf("%lu RFCOMM Channel %c: 0x%02.2x|%01.1x C/R: 0x%02.2x FrameType: 0x%02.2x P/F: 0x%02.2x\r\n",
    GetTickCount(),
    connected ? 'C' : '-',
    rfcommChannel >> 3,
    rfcommDirection >> 2,
    rfcommCommandResponse >> 1,
    rfcommFrameType,
    rfcommPfBit
  );)

  pBtd->hci_host_number_of_packets_completed(1);

  switch ( rfcommFrameType )
  {/*{{{*/
    case RFCOMM_DM: // Disconnected Mode Response
      // The remote side accepts no commands until the DM state is terminated by 
      // receipt of a SABM command. (TS 07.10 5.3.3)
      TRACE_USBHOST(printf("%lu Received DM Resp\r\n", GetTickCount());)
      break;
    case RFCOMM_SABM: // Set Asynchronous Unbalanced Mode
      // See 5.4.1, "DLC Establishment" (TS 07.10)
      // Respond with UA response; set DLC send/receive state variables to zero.
      // If ready to establish conn: 
      // - Notify initiating TE with UA frame having F-bit set to 1.
      // Otherwise
      // - Notify initiating TE with DM frame having F-bit set to 1.
      rfcommPfBit |= 0x10;
      TRACE_USBHOST(printf("%lu Received SABM Command\r\n", GetTickCount());)
      // This is ridiculous: The enqueue_sendrfcomm method merely
      // reassembles the bit fields extracted above.
      pBtd->enqueue_sendRfcomm(
        getServiceId(),
        r->handle,
        rfcomm_scid,
        rfcommChannel,
        rfcommDirection,
        rfcommCommandResponse,
        RFCOMM_UA, // Unnumbered Acknowledgement
        rfcommPfBit,
        rfcommbuf,
        0x00
      ); // UA Command
      break;
    case RFCOMM_UIH: // Unnumbered Information with Header Check 
      if (connected) {/*{{{*/
        if (rfcommChannel == rfcommChannelConnection) {/*{{{*/
          uint8_t length = r->length >> 1; // Get length
          uint8_t offset = l2capinbuf[4] - length - 4; // See if there is credit
          if (rfcommAvailable + length <= sizeof (rfcommDataBuffer)) {/*{{{*/
            // Don't add data to buffer if it would be full
            for (uint8_t i = 0; i < length; i++)
              rfcommDataBuffer[rfcommAvailable + i] = l2capinbuf[11 + i + offset];
            rfcommAvailable += length;
          }/*}}}*/
          /*{{{*/
#ifdef EXTRADEBUG
          TRACE_USBHOST(printf("%lu RFCOMM Data Available: %d\r\n", GetTickCount(),rfcommAvailable);)
          if (offset) {
            TRACE_USBHOST(printf("%lu  - Credit: 0x%02.2x", GetTickCount(),r->command);)
          }
#endif
          /*}}}*/
#ifdef PRINTREPORT // Uncomment "#define PRINTREPORT" to print the report send to the Arduino via Bluetooth
          for (uint8_t i = 0; i < length; i++) {
            TRACE_USBHOST(printf("%lu %02.2x", GetTickCount(),l2capinbuf[11+i+offset]);)
              TRACE_USBHOST(printf("%lu  ", GetTickCount());)
          }
#endif
        }/*}}}*/
        else if (r->command == BT_RFCOMM_RPN_CMD) {/*{{{*/
          // UIH Remote Port Negotiation Command
          TRACE_USBHOST(printf("%lu Received UIH Remote Port Negotiation Command B\r\n", GetTickCount());)
          rfcommbuf[0] = BT_RFCOMM_RPN_RSP; // Command
          // Echo packet contents back to remote TE
          memmove(&(rfcommbuf[1]), &(l2capinbuf[12]), 9);
          pBtd->enqueue_sendRfcomm(
            getServiceId(),
            r->handle,
            rfcomm_scid,
            rfcommChannel,
            rfcommDirection,
            0,
            RFCOMM_UIH,
            rfcommPfBit,
            rfcommbuf,
            0x0A
          ); // UIH Remote Port Negotiation Response
        }/*}}}*/
        else if (r->command == BT_RFCOMM_MSC_CMD) {/*{{{*/
          // UIH Modem Status Command
          TRACE_USBHOST(printf("%lu Send UIH Modem Status Response\r\n", GetTickCount());)
          rfcommbuf[0] = BT_RFCOMM_MSC_RSP; // UIH Modem Status Response
          rfcommbuf[1] = 2 << 1 | 1; // Length and shiftet like so: length << 1 | 1
          rfcommbuf[2] = l2capinbuf[13]; // Channel: (1 << 0) | (1 << 1) | (0 << 2) | (channel << 3)
          rfcommbuf[3] = l2capinbuf[14];
          pBtd->enqueue_sendRfcomm(
            getServiceId(),
            r->handle,
            rfcomm_scid,
            rfcommChannel,
            rfcommDirection,
            0,
            RFCOMM_UIH,
            rfcommPfBit,
            rfcommbuf,
            0x04
          );
        }/*}}}*/
      }/*}}}*/
      else
        switch ( r->command )
        {/*{{{*/
          case BT_RFCOMM_PN_CMD:
            {/*{{{*/
              // UIH Parameter Negotiation Command
              TRACE_USBHOST(printf("%lu Received UIH Parameter Negotiation Command\r\n", GetTickCount());)
              rfcommbuf[0] = BT_RFCOMM_PN_RSP; // UIH Parameter Negotiation Response
              rfcommbuf[1] = l2capinbuf[12]; // Length and shifted like so: length << 1 | 1
              rfcommbuf[2] = l2capinbuf[13]; // Channel: channel << 1 | 1
              // CL1-CL4
              rfcommbuf[3] = 0xE0; // Predefined for Bluetooth, see 5.5.3 of GSMTS 07.10
              rfcommbuf[4] = 0x00; // Priority
              rfcommbuf[5] = 0x00; // Timer
              // Octets 6-7 should be set to the maximum bulk packet size
              // minus header length 
              rfcommbuf[6] = ((L2CAP_IN_BUFFERSIZE - 14) % 0x100) & 0xff; // Max Fram Size LSB
              rfcommbuf[7] = ((L2CAP_IN_BUFFERSIZE - 14) / 0x100) & 0xff; // Max Fram Size MSB
              rfcommbuf[8] = 0x00;
              rfcommbuf[9] = 0x00;
              pBtd->enqueue_sendRfcomm(
                getServiceId(),
                r->handle,
                rfcomm_scid,
                rfcommChannel,
                rfcommDirection,
                0,
                RFCOMM_UIH,
                rfcommPfBit,
                rfcommbuf,
                0x0A
              );
            }/*}}}*/
            break;
          case BT_RFCOMM_MSC_CMD:
            {/*{{{*/
              // UIH Modem Status Command
              TRACE_USBHOST(printf("%lu Enqueue UIH Modem Status Response\r\n", GetTickCount());)
              rfcommbuf[0] = BT_RFCOMM_MSC_RSP; // UIH Modem Status Response
              rfcommbuf[1] = 2 << 1 | 1; // Length 
              rfcommbuf[2] = l2capinbuf[13]; // Channel
              rfcommbuf[3] = l2capinbuf[14];
              pBtd->enqueue_sendRfcomm(
                getServiceId(),
                r->handle,
                rfcomm_scid,
                rfcommChannel,
                rfcommDirection,
                0,
                RFCOMM_UIH,
                rfcommPfBit,
                rfcommbuf,
                0x04
              );

              TRACE_USBHOST(printf("%lu Enqueue UIH Modem Status Command\r\n", GetTickCount());)
              rfcommbuf[0] = BT_RFCOMM_MSC_CMD; // UIH Modem Status Command
              rfcommbuf[1] = 2 << 1 | 1; // Length
              rfcommbuf[2] = l2capinbuf[13]; // Channel
              rfcommbuf[3] = 0x8D; // Can receive frames (YES), Ready to Communicate (YES), Ready to Receive (YES), Incomig Call (NO), Data is Value (YES)
              pBtd->enqueue_sendRfcomm(
                getServiceId(),
                r->handle,
                rfcomm_scid,
                rfcommChannel,
                rfcommDirection,
                0,
                RFCOMM_UIH,
                rfcommPfBit,
                rfcommbuf,
                0x04
              );
            }/*}}}*/
            break;
          case BT_RFCOMM_MSC_RSP:
            {/*{{{*/
              // UIH Modem Status Response
              if (!creditSent) {
                // Enqueue a command to send flow control credit,
                // setting state flags after successful completion.
                TRACE_USBHOST(printf("%lu UIH command, sent flow control credit\r\n", GetTickCount());)
                pBtd->enqueue_sendRfcommCredit(
                  getServiceId(), 
                  r->handle,
                  rfcomm_scid,
                  rfcommChannelConnection,
                  rfcommDirection,
                  0,
                  RFCOMM_UIH,
                  0x10,
                  sizeof(rfcommDataBuffer),
                  RFCOMM_PROTOCOL_CREDIT_SENT 
                );
              }
            }/*}}}*/
            break;
          case BT_RFCOMM_RPN_CMD:
            {/*{{{*/
              // UIH Remote Port Negotiation Command
              TRACE_USBHOST(printf("%lu Received UIH Remote Port Negotiation Command A\r\n", GetTickCount());)
              rfcommbuf[0] = BT_RFCOMM_RPN_RSP; // Command
              // Copy length, channel, priority, timer, frame size, max retrans, and number of frames 
              memmove(&(rfcommbuf[1]), &(l2capinbuf[12]), 9);
              // Enqueue UIH Remote Port Negotiation Response
              pBtd->enqueue_sendRfcomm(
                getServiceId(),
                r->handle,
                rfcomm_scid,
                rfcommChannel,
                rfcommDirection,
                0,
                RFCOMM_UIH,
                rfcommPfBit,
                rfcommbuf,
                0x0A,
                RFCOMM_PROTOCOL_MESSAGE_SENT 
              );
              connected = true;
            }/*}}}*/
            break;
          default:
            if (r->length == 0x01) {/*{{{*/
              // UIH Command with credit
              TRACE_USBHOST(printf("%lu Received UIH flow control credit\r\n", GetTickCount());)
            }/*}}}*/
        }/*}}}*/
      break;
    case RFCOMM_DISC:
      {/*{{{*/
        TRACE_USBHOST(printf("%lu Received Disconnect RFCOMM Command on channel: 0x%02.2x\r\n", GetTickCount(),rfcommChannel >> 3);)
        connected = false;
        pBtd->enqueue_sendRfcomm(
          getServiceId(),
          r->handle,
          rfcomm_scid,
          rfcommChannel,
          rfcommDirection,
          rfcommCommandResponse,
          RFCOMM_UA,
          rfcommPfBit,
          rfcommbuf,
          0x00
        ); // UA Command
      }/*}}}*/

      break;

    case RFCOMM_UI: // Unnumbered Information - Unsupported in RFCOMM
    default:
      // rfcommFrameType != RFCOMM_DISC
      dump_frame = true;
      printf("%lu Unsupported RFCOMM Data - FrameType: 0x%02.2x Command: 0x%02.2x\r\n", 
        GetTickCount(),
        rfcommFrameType,
        r->command
      );
      break;
  }/*}}}*/

  if ( dump_frame ) {/*{{{*/
    for (uint16_t i = 0; i < L2CAP_IN_BUFFERSIZE; i++) {
      if (!(i % 16)) printf("\r\n%lu ", GetTickCount());
      if (!(i % 8)) printf(" ");
      printf(" %02.2x", l2capinbuf[i]);
    }
    printf("\r\n");
  }/*}}}*/

  return true;

}/*}}}*/

void SPP::ACLData(uint8_t *l2capinbuf) {/*{{{*/

  l2cap_message_t * p = (l2cap_message_t *)l2capinbuf;

  if ( p->channel_id == 0x0001 )
    handle_signaling_channel(l2capinbuf);
  else if ( p->channel_id == *((uint16_t *)&sdp_dcid) )
    handle_sdp_channel(l2capinbuf);
  else if ( p->channel_id == *((uint16_t *)&rfcomm_dcid) )
    handle_rfcomm_channel(l2capinbuf);
  else
  {/*{{{*/
    if ( p->channel_id != 0 ) {
      TRACE_USBHOST(printf("%lu ~~ Unsupported L2CAP Signaling Message, Channel ID: 0x%04.4X\r\n", GetTickCount(), p->channel_id);)
      TRACE_USBHOST(dump_l2cap_message_t(p);)
    }
    memset(l2capinbuf,0,L2CAP_IN_BUFFERSIZE);
    return;
  }/*}}}*/

  SDP_task();
  RFCOMM_task();
  // Clear inbound message buffer
  memset(l2capinbuf,0,L2CAP_IN_BUFFERSIZE);

}/*}}}*/

bool SPP::Run(uint8_t *l2capinbuf) {/*{{{*/

  // This method enqueues L2CAP and Bluetooth protocol communication commands
  // See SDP_task and RFCOMM_task queue methods
  ACLData(l2capinbuf);
	send();
}/*}}}*/

void SPP::StateChange(uint32_t f) {/*{{{*/

  // Invoked after sending protocol commands on bulk output
  // pipe (L2CAP, RFCOMM).
  //
  // Perform an internal state change action (changing flags, updating 
  // data buffers, etc.) based on a flag value
  //
  // Because these state changes are uncoupled from 
  // preceding commands in the blocking context, the 
  // original SDP_task and RFCOMM_task methods must be
  // modified to retain intended behavior.
  bool state_changed = true;
  if ( f > 0 ) {
    switch (f) {
      case L2CAP_CONFIG_REQUEST_SENT:
        // L2CAP SDP Configuration Request sent.
        l2cap_sdp_state = L2CAP_SDP_REQUEST;
        l2cap_event_flag |= L2CAP_FLAG_CONFIG_SDP_REQUEST;
        TRACE_USBHOST(printf("+ SDP Incoming Connection Request\r\n");)
        break;
      case L2CAP_CONFIG_RESPONSE_SENT:
        TRACE_USBHOST(printf("+ L2CAP Configuration Response Sent; Sending Configuration Request.\r\n");)
        break;

      case SDP_DISCOVERY_CONNECTIONLESS:
        printf("+ Connectionless Discovery Response Sent\r\n");
        break;
      case RFCOMM_PROTOCOL_CREDIT_SENT:
        creditSent = true;
        timer = GetTickCount();
        waitForLastCommand = true;
        break;

      case RFCOMM_PROTOCOL_MESSAGE_SENT:
        waitForLastCommand = false;
        creditSent = false;
        connected = true; // The RFCOMM channel is now established
        spp_available_out = 0;
        TRACE_USBHOST(printf("+ RFCOMM Connection is now established\r\n\r\n");)
        break;


      /** Vetted state handlers */
      case SDP_DISCOVERY_MESSAGETOGGLE:
        firstMessage = !firstMessage;
        TRACE_USBHOST(printf("+ Toggle firstMessage %c\r\n", firstMessage ? 'true' : 'false');)
        break;

      case SDP_CONN_INCREMENT_CONNID:
        l2cap_sdp_state = L2CAP_SDP_CONFIG_REQUEST;
        identifier++;
        TRACE_USBHOST(printf("+ 0x%02.2X Incremented connection identifier to \r\n", f);)
        break;

      case SDP_CONN_DISCONNECT_SDP_REQUEST:
				SDPConnected = false;
        l2cap_sdp_state = L2CAP_SDP_WAIT;
        l2cap_event_flag |= L2CAP_FLAG_DISCONNECT_RESPONSE;
        printf("+ Disconnected SDP Channel\r\n");
        break;

      case RFCOMM_CONN_DISCONNECT_REQUEST:
        l2cap_event_flag |= L2CAP_FLAG_DISCONNECT_RFCOMM_REQUEST;
				RFCOMMConnected = false;
				connected = false;
				l2cap_rfcomm_state = L2CAP_RFCOMM_WAIT;
				TRACE_USBHOST(printf("+ Disconnected RFCOMM Channel\r\n");)
        break;

      case RFCOMM_CONN_INCREMENT_CONNID:
        identifier++;
        l2cap_event_flag |= L2CAP_FLAG_CONFIG_RFCOMM_REQUEST;
        l2cap_rfcomm_state = L2CAP_RFCOMM_REQUEST;
        TRACE_USBHOST(printf("+ 0x%02.2X Incremented transaction ID, request config.\r\n", f);)
        break;

      case RFCOMM_CONN_L2CAP_RFCOMM_REQUEST:
        l2cap_rfcomm_state = L2CAP_RFCOMM_REQUEST;
        TRACE_USBHOST(printf("+ RFCOMM Incoming Connection Request\r\n");)
        break;

      case RFCOMM_CONN_RECEIVED_CONFIG_REPLY:
        l2cap_rfcomm_state = L2CAP_RFCOMM_DONE;
				rfcommAvailable = 0; // Reset number of bytes available
				bytesRead = 0; // Reset number of bytes received
				RFCOMMConnected = true;
        TRACE_USBHOST(printf("+ RFCOMM Successfully Configured\r\n");)
        break;

      case RFCOMM_CONN_SEND_COMPLETE:
        break;

      default:
        state_changed = false;
    }
    if ( state_changed ) {
      SDP_task();
    }
  }
  else SDP_task();
}/*}}}*/

void SPP::SDP_task() {/*{{{*/
  // Consult Core Specification V4.0, Core System Package BR/EDF,
  // Message Sequence Charts.
	switch (l2cap_sdp_state) {
		case L2CAP_SDP_WAIT:
			break;
    case L2CAP_SDP_CONFIG_REQUEST:
      break;
		case L2CAP_SDP_REQUEST:
			break;
		case L2CAP_SDP_SUCCESS:
			break;
		case L2CAP_SDP_DONE:
			break;
		case L2CAP_DISCONNECT_RESPONSE:
			break;
	}
}/*}}}*/

void SPP::RFCOMM_task() {/*{{{*/
  if ( pBtd->rfcommConnectionClaimed )
	switch (l2cap_rfcomm_state) {
		case L2CAP_RFCOMM_WAIT:
			break;
		case L2CAP_RFCOMM_REQUEST:
			break;
		case L2CAP_RFCOMM_SUCCESS:
			break;
		case L2CAP_RFCOMM_DONE:
			break;
    case L2CAP_RFCOMM_HOLD:
      break;
	}
}/*}}}*/

/* Serial commands */

size_t SPP::write(uint8_t data) {/*{{{*/
	return write(&data,1);
}/*}}}*/

size_t SPP::write(const uint8_t* data, size_t size) {/*{{{*/
	for(uint8_t i = 0; i < size; i++) {
		if(spp_available_out >= sizeof(sppOutputBuffer)/sizeof(sppOutputBuffer[0]))
			send(); // Send the current data in the buffer
		sppOutputBuffer[spp_available_out++] = data[i]; // All the bytes are put into a buffer and then send using the send() function
	}
	return size;
}/*}}}*/

int SPP::available(void) {/*{{{*/
	return rfcommAvailable;
}/*}}}*/

void SPP::flush(void) {/*{{{*/
	rfcommAvailable = 0;
}/*}}}*/

int SPP::peek(void) {/*{{{*/
	if (rfcommAvailable == 0) // Don't read if there is nothing in the buffer
		return -1;
	return rfcommDataBuffer[0];
}/*}}}*/

int SPP::read(void) {/*{{{*/
	if (rfcommAvailable == 0) // Don't read if there is nothing in the buffer
		return -1;
	uint8_t output = rfcommDataBuffer[0];
  memmove(rfcommDataBuffer, &(rfcommDataBuffer[1]), sizeof(rfcommDataBuffer) - 1);
	rfcommAvailable--;
	bytesRead++;
	if (bytesRead > (sizeof (rfcommDataBuffer) - 5)) { // We will send the command just before it runs out of credit
		bytesRead = 0;
		pBtd->enqueue_sendRfcommCredit(getServiceId(), hci_handle, rfcomm_scid, rfcommChannelConnection, rfcommDirection, 0, RFCOMM_UIH, 0x10, sizeof (rfcommDataBuffer)); // Send more credit
#ifdef EXTRADEBUG
		TRACE_USBHOST(printf("Sent %d more credit\r\n",(uint8_t)sizeof (rfcommDataBuffer));)
#endif
	}
	return output;
}/*}}}*/

void SPP::send() {/*{{{*/

	uint8_t length; // This is the length of the string we are sending

	if (!connected || !spp_available_out) {
		return;
	}

  // Check if the string is larger than the outgoing buffer
  length = (spp_available_out > sizeof(rfcommbuf))
    ? sizeof(rfcommbuf)
    : spp_available_out;

  // TODO: Use a ring buffer
  memmove(&(rfcommbuf[0]), &(sppOutputBuffer[0]), length);
  // Shift remaining bytes in buffer into zeroth position 
  memmove(&(sppOutputBuffer[0]), &(sppOutputBuffer[length]), sizeof(sppOutputBuffer) - length);

  pBtd->enqueue_sendRfcomm(
    getServiceId(),
    hci_handle,
    rfcomm_scid,
    // rfcommChannelConnection | 0 | 0 | extendAddress;
    rfcommChannelConnection,
    0, // rfcommDirection,
    0,
    // RFCOMM_UIH; // RFCOMM Control
    RFCOMM_UIH,
    0, // rfcommPfBit,
    rfcommbuf,
    length,
    RFCOMM_CONN_SEND_COMPLETE
  );

  spp_available_out -= length;
}/*}}}*/


