/* Copyright (C) 2011 Circuits At Home, LTD. All rights reserved.

   This software may be distributed and modified under the terms of the GNU
   General Public License version 2 (GPL2) as published by the Free Software
   Foundation and appearing in the file GPL2.TXT included in the packaging of
   this file. Please note that GPL2 Section 2[b] requires that all works based
   on this software must also be made publicly available under the terms of
   the GPL2 ("Copyleft").

   Contact information
   -------------------

   Circuits At Home, LTD
Web      :  http://www.circuitsathome.com
e-mail   :  support@circuitsathome.com
*/

/* USB functions */

#include "Arduino.h"
#include "Usb.h"
#include <stdio.h>

#define LEGACY_INITIALIZATION
// #define DEBUG_USBHOST_USERLAND_EVENTS

volatile static uint32_t usb_error = 0;
volatile static uint32_t usb_task_state = USB_DETACHED_SUBSTATE_INITIALIZE;

/** ASF callbacks **/

void main_usb_reset_callback(void) {/*{{{*/
  if ( usb_task_state == USB_ATTACHED_SUBSTATE_WAIT_RESET_COMPLETE ) {
    TRACE_USBHOST(printf(" + USB_ATTACHED_SUBSTATE_WAIT_RESET_COMPLETE\r\n");)
    usb_task_state = USB_ATTACHED_SUBSTATE_WAIT_SOF;
    uhd_enable_sof();
  }
}/*}}}*/

void main_usb_mode_change(bool s) {/*{{{*/
  printf("~~ Mode %c\r\n", s ? 'H' : 'D');
}/*}}}*/

void main_usb_vbus_change(bool b_present) {/*{{{*/
#ifdef DEBUG_USBHOST_USERLAND_EVENTS
  printf("~~ VBus %c\r\n", b_present ? '1' : '0');
#endif
}/*}}}*/

void main_usb_vbus_error() {/*{{{*/
  printf("~~ VBus error\r\n");
  if ( UHD_GetVBUSState() == UHD_STATE_DISCONNECTED ) {
    usb_task_state = USB_DETACHED_SUBSTATE_INITIALIZE;
  }
}/*}}}*/

void main_usb_connection_event(uhc_device_t *dev, bool b_present) {/*{{{*/
#ifdef DEBUG_USBHOST_USERLAND_EVENTS
  printf("~~ MConn %c\r\n", b_present ? '1' : '0');
#endif
}/*}}}*/

void main_usb_wakeup_event() {/*{{{*/
  printf("~~ USB wakeup\r\n");
}/*}}}*/

void main_usb_sof_event() {/*{{{*/
  if ( usb_task_state == USB_ATTACHED_SUBSTATE_WAIT_SOF ) {
    TRACE_USBHOST(printf(" + USB_ATTACHED_SUBSTATE_WAIT_SOF\r\n");)
    usb_task_state = USB_STATE_CONFIGURING;
  }
}/*}}}*/

void main_usb_enum_event(uhc_device_t *dev, uhc_enum_status_t state)
{/*{{{*/
#ifdef DEBUG_USBHOST_USERLAND_EVENTS
  printf("~~ Enum %u (0x%02.2X)\r\n", (uint16_t)state, (uint8_t)state);
#endif
  if ( state == UHC_ENUM_SUCCESS ) {
    if ( usb_task_state == USB_STATE_CONFIGURING ) {
      usb_task_state = USB_STATE_CONFIGURING_ACTIVE;
    }
  }
  else {
    // Enumeration failed
    usb_task_state = USB_ATTACHED_SUBSTATE_SETTLE;
  }
}/*}}}*/

void main_usb_vendor_change(uhc_device_t *dev, bool b_status)
{/*{{{*/
#ifdef DEBUG_USBHOST_USERLAND_EVENTS
  printf("~~ Change %c\r\n", b_status ? '1' : '0');
#endif
}/*}}}*/


/**
 * \brief USBHost class constructor.
 */
USBHost::USBHost() : bmHubPre(0), hci_state(0), retries(0), nak(false), err(false) 
{/*{{{*/
  // Set up state machine
  usb_task_state = USB_DETACHED_SUBSTATE_INITIALIZE;

  // Init host stack
  for ( uint32_t i = 0 ; i < USB_NUMDEVICES ; i++ )
    devConfig[i] = 0;

  init();

}/*}}}*/

/**
 * \brief Initialize USBHost class.
 */
void USBHost::init()
{/*{{{*/
  devConfigIndex = 0;
  bmHubPre       = 0;
}/*}}}*/


/**
 * \brief Get USBHost state.
 *
 * \return USB enumeration status (see USBHost::task).
 */
uint32_t USBHost::getUsbTaskState(void)
{/*{{{*/
  return (usb_task_state);
}/*}}}*/

/**
 * \brief Set USB state.
 *
 * \param state New USBHost status to be set.
 */
void USBHost::setUsbTaskState(uint32_t state)
{/*{{{*/
  usb_task_state = state;
}/*}}}*/

/**
 * \brief Get HCI state.
 *
 * \param state New USBHost status to be set.
 */
void USBHost::setHciState(uint32_t hci_state_new)
{/*{{{*/
  hci_state = hci_state_new; 
}/*}}}*/

uint32_t USBHost::getHciState(void)
{/*{{{*/
  return (hci_state); 
}/*}}}*/

bool USBHost::getNakFlag(void)
{/*{{{*/
  return (nak); 
}/*}}}*/

bool USBHost::getErrFlag(void)
{/*{{{*/
  return (err); 
}/*}}}*/

bool USBHost::setErrFlag(bool err_)
{/*{{{*/
  err = err_;
}/*}}}*/


/**
 * \brief Get endpoint info from USB device address and device endpoint.
 *
 * \note This function should be used to know which host pipe is being used for
 * the corresponding device endpoint.
 *
 * \param addr USB device address.
 * \param ep USB device endpoint number.
 *
 * \return Pointer to an EpInfo structure.
 */
EpInfo* USBHost::getEpInfoEntry(uint32_t addr, uint32_t ep)
{/*{{{*/
  UsbDevice *p = addrPool.GetUsbDevicePtr(addr);

  if (!p || !p->epinfo) {
    printf("! No UsbDevice * matches addr %u, endpoint 0x%02.2X\r\n", addr, ep);
    return NULL;
  }

  EpInfo *pep = p->epinfo;

  for (uint32_t i = 0; i < p->epcount; i++)
  {
    if ((uint32_t)(pep->deviceEpNum & 0xff) == (uint32_t)(ep & 0x00ff)) {
      return pep;
    }
    pep++;
  }

  printf("! No matches for addr %u, endpoint 0x%02.2X in set of %u\r\n", addr, ep, p->epcount);

  return NULL;
}/*}}}*/

/**
 * \brief Set device endpoint entry.
 *
 * \note Each device is different and has a different number of endpoints.
 * This function sets endpoint record structure to the device using address
 * addr in the address pool.
 *
 * \param ul_pipe Pipe address.
 * \param ul_token_type Token type.
 *
 * \retval 0 on success.
 * \retval USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL device not found.
 */
uint32_t USBHost::setEpInfoEntry(uint32_t addr, uint32_t epcount, EpInfo* eprecord_ptr)
{/*{{{*/
  if (!eprecord_ptr)
    return USB_ERROR_INVALID_ARGUMENT;

  UsbDevice *p = addrPool.GetUsbDevicePtr(addr);

  if (!p) {
    printf("/!\\ USBHost::setEpInfoEntry : UsbDevice pointer NULL addr %u\r\n", addr);
    return USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL;
  }

  p->address = addr;
  p->epinfo  = eprecord_ptr;
  p->epcount = epcount;

  return 0;
}/*}}}*/

/**
 * \brief Configure device using known device classes.
 * The device get a new address even if its class remain unknown.
 *
 * \param parent USB device address of the device's parent (0 if root).
 * \param port USB device base address (see AddressPoolImpl).
 * \param lowspeed Device speed.
 *
 * \return 0 on success, error code otherwise.
 */
uint32_t USBHost::Configuring(uint32_t parent, uint32_t port, uint32_t lowspeed)
{/*{{{*/
  uint32_t rcode = 0;

  for (; devConfigIndex < USB_NUMDEVICES; ++devConfigIndex)
  {
    if (!devConfig[devConfigIndex])
      continue;

    rcode = devConfig[devConfigIndex]->Init(parent, port, lowspeed);

    if (!rcode) {
      TRACE_USBHOST(printf("USBHost::Configuring : found device class!\r\n");)
      devConfigIndex = 0;
      return 0;
    }

    if (rcode == USB_DEV_CONFIG_ERROR_DEVICE_NOT_SUPPORTED)
    {
      TRACE_USBHOST(printf("USBHost::Configuring : ERROR : device not supported!\r\n");)
    }
    else if (rcode == USB_ERROR_CLASS_INSTANCE_ALREADY_IN_USE)
    {
      TRACE_USBHOST(printf("USBHost::Configuring : ERROR : class instance already in use!\r\n");)
    }
    else
    {
      // in case of an error devConfigIndex should be reset to 0
      // in order to start from the very beginning the next time
      // the program gets here
      if (rcode != USB_DEV_CONFIG_ERROR_DEVICE_INIT_INCOMPLETE)
        devConfigIndex = 0;

      return rcode;
    }
  }

  // Device class is not supported by any of the registered classes
  devConfigIndex = 0;

  return 0;
}/*}}}*/

/**
 * \brief Release device and free associated resources.
 *
 * \param addr USB device address.
 *
 * \return 0 on success, error code otherwise.
 */
uint32_t USBHost::ReleaseDevice(uint32_t addr)
{/*{{{*/
  if (!addr)
    return 0;

  for (uint32_t i = 0; i < USB_NUMDEVICES; ++i)
  {
    if (devConfig[i]->GetAddress() == addr)
    {
      return devConfig[i]->Release();
    }
  }

  return 0;
}/*}}}*/

/**
 * \brief Get device descriptor.
 *
 * \param addr USB device address.
 * \param ep USB device endpoint number.
 * \param nbytes Buffer size.
 * \param dataptr Buffer to store received descriptor.
 *
 * \return 0 on success, error code otherwise.
 */
uint32_t USBHost::getDevDescr(uint32_t addr, uint32_t ep, uint32_t nbytes, uint8_t* dataptr)
{/*{{{*/
  TRACE_USBHOST(printf("   => USBHost::getDevDescr(%u,%u,%u...) [user buffer]\r\n",addr,ep,nbytes);)
  return (ctrlReq(
    addr                      , // USB Device address
    ep                        , // ep USB device endpoint number.
    bmREQ_GET_DESCR           , // bmReqType Request direction.
    USB_REQUEST_GET_DESCRIPTOR, // bRequest Request type.
    0x00                      , // wValLo Value low.
    USB_DESCRIPTOR_DEVICE     , // wValHi Value high.
    0x0000                    , // wInd Index field.
    nbytes                    , // total Request length.
    nbytes                    , // nbytes Number of bytes to read.
    dataptr                   , // dataptr Data pointer.
    0                           // p USB class reader.
  ));
}/*}}}*/

/**
 * \brief Get configuration descriptor.
 *
 * \param addr USB device address.
 * \param ep USB device endpoint number.
 * \param nbytes Buffer size.
 * \param conf Configuration number.
 * \param dataptr Buffer to store received descriptor.
 *
 * \return 0 on success, error code otherwise.
 */
uint32_t USBHost::getConfDescr(uint32_t addr, uint32_t ep, uint32_t nbytes, uint32_t conf, uint8_t* dataptr)
{/*{{{*/
  TRACE_USBHOST(printf("   => USBHost::getConfDescr(%u,%u,%u,%04.4x...) [user buffer]\r\n",addr,ep,nbytes,conf);)
  return (ctrlReq(addr, ep, bmREQ_GET_DESCR, USB_REQUEST_GET_DESCRIPTOR, conf, USB_DESCRIPTOR_CONFIGURATION, 0x0000, nbytes, nbytes, dataptr, 0));
}/*}}}*/

/**
 * \brief Get configuration descriptor and extract endpoints using USBReadParser object.
 *
 * \param addr USB device address.
 * \param ep USB device endpoint number.
 * \param conf Configuration number.
 * \param p USBReadParser object pointer used to extract endpoints.
 *
 * \return 0 on success, error code otherwise.
 */
uint32_t USBHost::getConfDescr(uint32_t addr, uint32_t ep, uint32_t conf, USBReadParser *p)
{/*{{{*/
  const uint32_t bufSize = 256;
  uint8_t buf[bufSize];

  uint32_t ret = getConfDescr(addr, ep, 8, conf, buf);

  if (ret) {
    return ret;
  }

  uint32_t total = ((USB_CONFIGURATION_DESCRIPTOR*)buf)->wTotalLength;
  delay(100);

  TRACE_USBHOST(printf("   => USBHost::getConfDescr(%u,%u,%u/%u,0x%04.4x...) [fixed buffer]\r\n",addr,ep,total,bufSize,conf);)
  return (ctrlReq(addr, ep, bmREQ_GET_DESCR, USB_REQUEST_GET_DESCRIPTOR, conf, USB_DESCRIPTOR_CONFIGURATION, 0x0000, total, bufSize, buf, p));
}/*}}}*/

/**
 * \brief Get string descriptor.
 *
 * \param addr USB device address.
 * \param ep USB device endpoint number.
 * \param nbytes Buffer size.
 * \param index String index.
 * \param langid Language ID.
 * \param dataptr Buffer to store received descriptor.
 *
 * \return 0 on success, error code otherwise.
 */
uint32_t USBHost::getStrDescr(uint32_t addr, uint32_t ep, uint32_t nbytes, uint8_t index, uint16_t langid, uint8_t* dataptr)
{/*{{{*/
  TRACE_USBHOST(printf("   => USBHost::getStrDescr\r\n");)
  return (ctrlReq(addr, ep, bmREQ_GET_DESCR, USB_REQUEST_GET_DESCRIPTOR, index, USB_DESCRIPTOR_STRING, langid, nbytes, nbytes, dataptr, 0));
}/*}}}*/


/**
 * \brief USB main task, responsible for enumeration and clean up stage.
 *
 * \note Must be periodically called from loop().
 */
void USBHost::Task(void) {/*{{{*/

  uint32_t rcode = 0;
  static uint32_t delay = 0;
  uint32_t lowspeed = 0;

  // Update USB task state on Vbus change
  switch (UHD_GetVBUSState())
  {
    case UHD_STATE_ERROR:
      // Illegal state
      usb_task_state = USB_DETACHED_SUBSTATE_ILLEGAL;
      lowspeed = 0;
      err = true;
      break;

    case UHD_STATE_DISCONNECTED:
      // Disconnected state
      break;

    case UHD_STATE_CONNECTED:
      // Attached state
      if ((usb_task_state & USB_STATE_MASK) == USB_STATE_DETACHED)
      {
        delay = GetTickCount() + USB_SETTLE_DELAY;
        usb_task_state = USB_ATTACHED_SUBSTATE_SETTLE;
        //FIXME TODO: lowspeed = 0 ou 1;  already done by hardware?
      }
      break;

    case UHD_STATE_NO_VBUS:
      // Transient state following call to UHD_Init() 
      break;
  }

  // Poll connected devices (if required)
  for (uint32_t i = 0; i < USB_NUMDEVICES; ++i)
    if (devConfig[i])
      rcode = devConfig[i]->Poll();

  // Perform USB enumeration stage and clean up
  switch (usb_task_state)
  {
    case USB_DETACHED_SUBSTATE_INITIALIZE:
      TRACE_USBHOST(printf(" + USB_DETACHED_SUBSTATE_INITIALIZE\r\n");)

      // Init USB stack and driver
      init();

      // Free all USB resources
      for (uint32_t i = 0; i < USB_NUMDEVICES; ++i)
        if (devConfig[i])
          rcode = devConfig[i]->Release();

      usb_task_state = USB_DETACHED_SUBSTATE_WAIT_FOR_DEVICE;

      break;

    case USB_DETACHED_SUBSTATE_WAIT_FOR_DEVICE:
      // Nothing to do
      break;

    case USB_DETACHED_SUBSTATE_ILLEGAL:
      // Nothing to do
      break;

    case USB_ATTACHED_SUBSTATE_SETTLE:
      // Settle time for just attached device
      if (delay < GetTickCount())
      {
        TRACE_USBHOST(printf(" + USB_ATTACHED_SUBSTATE_SETTLE\r\n");)
          usb_task_state = USB_ATTACHED_SUBSTATE_RESET_DEVICE;
      }
      break;

    case USB_ATTACHED_SUBSTATE_RESET_DEVICE:
      TRACE_USBHOST(printf(" + USB_ATTACHED_SUBSTATE_RESET_DEVICE\r\n");)
      // Trigger Bus Reset
      usb_task_state = USB_ATTACHED_SUBSTATE_WAIT_RESET_COMPLETE;
      UHD_BusReset();
      delay = GetTickCount() + 20;
      break;

    case USB_ATTACHED_SUBSTATE_WAIT_RESET_COMPLETE:
      if (delay < GetTickCount()) {
        // Switch to USB_ATTACHED_SUBSTATE_WAIT_SOF
        // if we're stalled here without 
        // main_usb_reset_callback() being called.
#if 0
        // Clear Bus Reset flag
        uhd_ack_reset_sent(); // Handled by ASF
        // Enable Start Of Frame generation
        uhd_enable_sof(); // Invoked by callback
#endif
        usb_task_state = USB_ATTACHED_SUBSTATE_WAIT_SOF;

        // Wait 20ms after Bus Reset (USB spec)
        delay = GetTickCount() + 20;
      }
      break;

    case USB_ATTACHED_SUBSTATE_WAIT_SOF:
      if (delay < GetTickCount()) {
        // SOF wait overtime elapsed
        TRACE_USBHOST(printf(" + USB_ATTACHED_SUBSTATE_WAIT_SOF ++\r\n");)
        // 20ms waiting elapsed - reset
        usb_task_state = USB_ATTACHED_SUBSTATE_SETTLE;
        delay = GetTickCount() + 200;
      }
      else
        usb_task_state = USB_STATE_CONFIGURING;
      break;

    case USB_STATE_CONFIGURING:
      // Interrupt-driven enumeration in progress
      break;

    case USB_STATE_CONFIGURING_ACTIVE:
#if defined(LEGACY_INITIALIZATION)
      TRACE_USBHOST(printf(" + USB_STATE_CONFIGURING\r\n");)
      rcode = Configuring(0, 0, lowspeed);

      err = !(rcode == 0);

      if (rcode)
      {
        printf("/!\\ USBHost::Task : USB_STATE_CONFIGURING failed with code: %lu\r\n", rcode);
          if (rcode != USB_DEV_CONFIG_ERROR_DEVICE_INIT_INCOMPLETE)
          {
            usb_error = rcode;
            usb_task_state = USB_STATE_ERROR;
            delay = GetTickCount() + 200;
          }
      }
      else
      {
        usb_task_state = USB_STATE_RUNNING;
        TRACE_USBHOST(printf(" + USB_STATE_RUNNING\r\n");)
      }
#else
      TRACE_USBHOST(printf(" + USB_STATE_CONFIGURING UNIMPLEMENTED\r\n");)
      usb_task_state = USB_STATE_RUNNING;
#endif
      break;

    case USB_STATE_RUNNING:
      break;

    case USB_STATE_ERROR:
      if (delay < GetTickCount()) {
        printf(" + USB_STATE_ERROR - Reinitializing\r\n");
        usb_task_state = USB_DETACHED_SUBSTATE_INITIALIZE;
      }
      break;
  }
}/*}}}*/

static bool transfer_request_complete      = false;
static uint32_t transfer_request_end_state = UHD_TRANS_NOERROR;
static uint16_t transfer_request_datasize  = 0;

/**
 * \brief Handle blocking Transfer result.
 *
 */
void transfer_result_callback(
    usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data) 
{/*{{{*/
  transfer_request_datasize = ( status == UHD_TRANS_NOERROR )
    ? (uint16_t)nb_transferred
    : 0
    ;
  transfer_request_end_state = status;
  printf("~+ Trans 0x%02.2X %s %u\r\n", ep, 
      status == UHD_TRANS_NOERROR ? "Ok" : "??", 
      transfer_request_datasize);
  transfer_request_complete = true;
}/*}}}*/

uint32_t USBHost::Transfer(uint32_t addr, uint32_t ep, uint32_t *nbytes, uint8_t* data)
{/*{{{*/

  // DEPRECATED: Use callback infrastructure to manage state transitions.
  // Use uhi_vendor.c calls
  // Execute a blocking transfer call, mimicking
  // behavior of original InTransfer / OutTransfer methods.
  uint32_t timeout = GetTickCount() + 20; 
  bool transfer_result;
 
  transfer_request_datasize  = 0;
  transfer_request_complete  = false;
  transfer_request_end_state = UHD_TRANS_TIMEOUT;

  transfer_result = uhd_ep_run(
    addr,
    ep, 
    false,
    (uint8_t *)data,
    *nbytes,
    0,
    &transfer_result_callback,
    0
  );

  if (!transfer_result) {
    return transfer_request_end_state;
  }

  while (!transfer_request_complete && (timeout > GetTickCount()));

  *nbytes = transfer_request_datasize;

  printf( "Done ex %llu state %s\r\n",
     *nbytes, transfer_request_end_state == UHD_TRANS_NOERROR ? "Ok" : "??" );

  return transfer_request_end_state;

}/*}}}*/

/**
 * \brief Send a control request.
 * Sets address, endpoint, fills control packet with necessary data, dispatches
 * control packet, and initiates bulk IN transfer depending on request.
 *
 * \param addr USB device address.
 * \param ep USB device endpoint number.
 * \param bmReqType Request direction.
 * \param bRequest Request type.
 * \param wValLo Value low.
 * \param wValHi Value high.
 * \param wInd Index field.
 * \param total Request length.
 * \param nbytes Number of bytes to read.
 * \param dataptr Data pointer.
 * \param p USB class reader.
 *
 * \return 0 on success, error code otherwise.
 */

static bool ctrl_request_complete     = false;
static uhd_trans_status_t ctrl_request_end_state;
static uint16_t ctrl_request_datasize = 0;

static void ctrlReq_response_handler(usb_add_t add,
		uhd_trans_status_t status,
		uint16_t payload_trans) {
  if ( status == UHD_TRANS_NOERROR ) {
    ctrl_request_datasize  = payload_trans;
    TRACE_USBHOST(printf("~ OK %u\r\n", payload_trans);)
  }
  else {
    ctrl_request_datasize  = 0;
    printf("~ NOK\r\n");
  }
  // Always set this flag to allow ctrlReq processing to continue.
  ctrl_request_complete = true;
  ctrl_request_end_state = status;
}
  
uint32_t USBHost::ctrlReq(uint32_t addr, uint32_t ep, uint8_t bmReqType, uint8_t bRequest, uint8_t wValLo, uint8_t wValHi, uint16_t wInd, uint16_t request_length, uint32_t recv_buf_length, uint8_t* dataptr, USBReadParser *p)
{/*{{{*/

  // Setup a control request using ASF framework
  // Return 0 on success, or an error code > 0 on failure.

  uint32_t timeout = GetTickCount() + 20; 
  usb_setup_req_t req;

  req.bmRequestType = bmReqType;
  req.bRequest      = bRequest;
  req.wValue        = (wValHi << 8) | wValLo;
  req.wIndex        = wInd;
  req.wLength       = request_length;  

  ctrl_request_complete  = false;
  ctrl_request_end_state = UHD_TRANS_TIMEOUT;

  if( !uhd_setup_request(
      addr,
      &req,
      (uint8_t *)dataptr,
      recv_buf_length,
      NULL,
      &ctrlReq_response_handler )) {
    printf("ctrlReq: Setup\r\n");
    return UHD_TRANS_ABORTED;
  } 

  while ( !ctrl_request_complete && (timeout > GetTickCount()) );

  if ( !ctrl_request_complete ) return UHD_TRANS_TIMEOUT;

  if ( ( bmReqType & 0x80 ) > 0 && dataptr != 0 ) {
    // Inbound transfer
    if ( p && ctrl_request_datasize > 0 ) 
      ((USBReadParser*)p)->Parse(
        ctrl_request_datasize, 
        dataptr, request_length
      );
  }

  return 0;

}/*}}}*/


